import sys
from pathlib import Path

from loguru import logger


def set_basic_handler(filename: str):
    return [
        {
            "sink": sys.stdout,
            "format": "{message}",
            "level": "ERROR",
        },
        {
            "sink": f"logs/{filename}.log",
            "format": "{message}",
            "level": "WARNING",
            "serialize": True,
        },
    ]


def set_info_handler(filename: str):
    return [
        {
            "sink": sys.stdout,
            "format": "{message}",
            "level": "DEBUG",
        },
        {
            "sink": f"logs/{filename}.log",
            "format": "{message}",
            "level": "INFO",
            "serialize": True,
        },
    ]


logger.configure(
    handlers=[
        {
            "sink": "logs/debug.log",
            "format": "{message}",
            "level": "DEBUG",
            "serialize": True,
        },
        {
            "sink": "logs/error.log",
            "format": "{message}",
            "level": "ERROR",
            "serialize": True,
        },
    ]
)


def clear_logs(p: Path = Path().cwd() / "logs"):
    for i in p.glob("*"):
        if i.is_file():
            i.unlink()



import abc
from pathlib import Path
from typing import Any

from corpus_pax import Individual
from jinja2 import Environment, PackageLoader, select_autoescape
from pydantic import BaseModel, EmailStr
from sqlpyd import Connection

BASE = "code/corpus"
"""Source of all local files"""

STATUTE_PATH = Path().home().joinpath(f"{BASE}/statutes")
STATUTE_FILES = STATUTE_PATH.glob("**/*/details.yaml")

CODIFICATION_PATH = Path().home().joinpath(f"{BASE}/codifications")
CODIFICATION_FILES = CODIFICATION_PATH.glob("**/*.yaml")

DOCUMENT_PATH = Path().home().joinpath(f"{BASE}/documents")
DOCUMENT_FILES = DOCUMENT_PATH.glob("**/*.yaml")

INCLUSION_FILE = "inclusions.yaml"
"""Name of the file that will host all the included components of a decision"""


corpus_sqlenv = Environment(
    loader=PackageLoader(package_name="corpus_x", package_path="sql"),
    autoescape=select_autoescape(),
)


class Integrator(BaseModel, abc.ABC):
    """
    Each Integrator class ensures that inheriting classes implement:

    1. common fields: `id`, `emails`, `meta`, `tree`, and `unit_fts`
    2. `make_tables()`: create the model's instances in the sqlite db
    3. `add_rows()`: to populate the tables created
    4. `from_page()`: given a raw yaml file, extract fields into the BaseModel
    5. `@relations`: the BaseModel will have relationships to other BaseModels

    The reason for requiring `emails` is that the common
    `insert_objects()` function can: create an m2m table with respect
    to authors

    The reason for requiring `@relations` is that the common
    `insert_objects()` function can: go through each of the tuples where in
    the first item of the tuple represents a table and the second item of the
    tuple, rows to be inserted in such a table.
    """

    id: str = NotImplemented
    emails: list[EmailStr] = NotImplemented
    meta: Any = NotImplemented
    tree: list[Any] = NotImplemented
    unit_fts: list[Any] = NotImplemented

    @classmethod
    @abc.abstractmethod
    def make_tables(cls, c: Connection) -> None:
        """Common process for creatng the tables associated
        with the concrete class."""
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def add_rows(cls, c: Connection) -> None:
        """Common process for creating objects from the source
        files for these to become prospective rows to
        the tables created."""
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def from_page(cls, file_path: Path) -> None:
        """The `file_path` expects an appropriate .yaml file
        containing the metadata. The data will be processed into an
        interim 'page' that will eventually build an instance of
        the concrete class."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def relations(cls):
        """Helper property to associate TableConfigured models
        to their instantiated values in preparation for
        database insertion."""
        raise NotImplementedError

    def insert_objects(
        self,
        c: Connection,
        obj: Any,
        correlations: list[tuple[Any, Any]],
    ) -> str:
        """The use of the concrete class' `insert_objects()` function
        implies that an `Individual` table already exists.

        The `obj` is a subclass of `TableConfig`. Since we're already aware
        of the `id` of the `obj`, we can also use this same id to create the
        author of the object as well as the correlated entities.

        Each correlated entity must also be a subclass of `TableConfig`.
        """
        record = self.meta.dict(exclude={"emails"})
        c.add_record(obj, record)

        for email in self.emails:
            c.table(obj).update(self.id).m2m(
                other_table=c.table(Individual),
                lookup={"email": email},
                pk="id",
            )

        for related in correlations:
            c.add_cleaned_records(
                related[0],  # the related model which must be
                related[1],  # the instance of the object
            )

        return self.id


def sql_get_detail(generic_tbl_name: str, generic_id: str) -> str:
    return corpus_sqlenv.get_template("base/get_detail.sql").render(
        generic_tbl=generic_tbl_name,
        target_id=generic_id,
    )


def sql_get_authors(generic_tbl_name: str, generic_id: str) -> str:
    """Produce the SQL query string necessary to get the authors from the
    Individual table based on the `generic_tbl_name`'s target `generic_id`.

    Each generic_tbl_name will be sourced from either: DecisionRow,
    CodeRow, DocRow, StatuteRow. Each of these tables are associated
    with the Individual table. The result looks something like this:

    ```
    >>> from .statutes import StatuteRow
    >>> sql = sql_get_authors(StatuteRow.__tablename__, "ra-386-june-18-1949")
    >>> type(sql)
    <class 'str'>
    ```

    See sqlite_utils which creates m2m object tables after sorting the tables
    alphabetically.
    """
    tables = [generic_tbl_name, Individual.__tablename__]
    template = corpus_sqlenv.get_template("base/get_author_ids.sql")
    return template.render(
        generic_tbl="_".join(sorted(tables)),
        col_generic_obj="_".join([generic_tbl_name, "id"]),
        col_author_id="_".join([Individual.__tablename__, "id"]),
        target_id=generic_id,
    )


def get_authored_object(
    c: Connection, generic_tbl_name: str, generic_id: str
) -> dict:
    tbl = generic_tbl_name
    idx = generic_id
    a = c.db.execute_returning_dicts(sql_get_detail(tbl, idx))[0]
    b = c.db.execute_returning_dicts(sql_get_authors(tbl, idx))[0]
    result = a | b
    return result



import sqlite3
from collections.abc import Iterator
from pathlib import Path
from sqlite3 import IntegrityError

from pydantic import EmailStr, Field, ValidationError
from sqlpyd import Connection, TableConfig
from statute_patterns import (
    Rule,
    StatuteSerialCategory,
    StatuteTitleCategory,
    extract_rules,
)
from statute_trees import (
    Node,
    Page,
    StatuteBase,
    StatutePage,
    StatuteUnit,
    generic_content,
    generic_mp,
)

from .resources import STATUTE_FILES, STATUTE_PATH, Integrator, corpus_sqlenv
from .utils import logger


class StatuteRow(Page, StatuteBase, TableConfig):
    """This corresponds to statute_trees.StatutePage but is adjusted
    for the purpose of table creation."""

    __prefix__ = "lex"
    __tablename__ = "statutes"
    __indexes__ = [
        ["statute_category", "statute_serial_id", "date", "variant"],
        ["statute_category", "statute_serial_id", "date"],
        ["statute_category", "statute_serial_id", "variant"],
        ["statute_category", "statute_serial_id"],
    ]

    @classmethod
    def get_base_data(cls, c: Connection, pk: str) -> dict:
        from .codifications import CodeRow

        sql_file = "statutes/get_base.sql"
        template = corpus_sqlenv.get_template(sql_file)
        results = c.db.execute_returning_dicts(
            template.render(
                target_statute_id=pk,
                statute_tbl=cls.__tablename__,
                statute_title_tbl=StatuteTitleRow.__tablename__,
                statute_references_tbl=StatuteFoundInUnit.__tablename__,
                code_tbl=CodeRow.__tablename__,
            )
        )
        if results:
            return results[0]
        return {}

    @classmethod
    def get_id_via_catid(cls, c: Connection, cat: str, id: str) -> str | None:
        tbl = c.table(cls)
        q = "statute_category = ? and statute_serial_id = ?"
        rows = list(tbl.rows_where(where=q, where_args=(cat, id), select="id"))
        idx = rows[0]["id"] if rows else None
        return idx

    @classmethod
    def get_id(cls, c: Connection, pk: str) -> str | None:
        tbl = c.table(cls)
        q = "id = ?"
        rows = list(tbl.rows_where(where=q, where_args=(pk,), select="id"))
        idx = rows[0]["id"] if rows else None
        return idx


class StatuteTitleRow(TableConfig):
    """This corresponds to statute_patterns.StatuteTitle but
    is adjusted for the purpose of table creation."""

    __prefix__ = "lex"
    __tablename__ = "statute_titles"
    __indexes__ = [["category", "text"], ["category", "statute_id"]]
    statute_id: str = Field(..., col=str, fk=(StatuteRow.__tablename__, "id"))
    category: StatuteTitleCategory = Field(
        ...,
        col=str,
        index=True,
    )
    text: str = Field(..., col=str, fts=True)

    class Config:
        use_enum_values = True


class StatuteUnitSearch(TableConfig):
    __prefix__ = "lex"
    __tablename__ = "statute_fts_units"
    __indexes__ = [["statute_id", "material_path"]]
    statute_id: str = Field(..., col=str, fk=(StatuteRow.__tablename__, "id"))
    material_path: str = generic_mp
    unit_text: str = generic_content


class StatuteMaterialPath(Node, TableConfig):
    __prefix__ = "lex"
    __tablename__ = "statute_mp_units"
    __indexes__ = [
        ["item", "caption", "content", "statute_id"],
        ["item", "caption", "statute_id"],
        ["item", "content", "statute_id"],
        ["item", "statute_id"],
    ]
    statute_id: str = Field(..., col=str, fk=(StatuteRow.__tablename__, "id"))
    material_path: str = generic_mp


class StatuteFoundInUnit(StatuteBase, TableConfig):
    """Each unit in Statute A (see MP) may refer to Statute B.
    Statute B is referenced through it's category and identifier
    (hence inheriting from `StatuteBase`). After securing the category
    and identifier pairs, can use the `cls.update_statute_ids()` to
    supply the matching statute  id of the category/identifier pair.
    """

    __prefix__ = "lex"
    __tablename__ = "statute_unit_references"
    __indexes__ = [
        ["statute_category", "statute_serial_id"],
        ["statute_category", "statute_id"],
    ]
    statute_id: str = Field(..., col=str, fk=(StatuteRow.__tablename__, "id"))
    material_path: str = generic_mp
    matching_statute_id: str | None = Field(
        None,
        description=(
            "Each unit in Statute A (see MP) may refer to Statute B."
            " Statute B is referenced through it's category and identifier"
            " (see StatuteBase)."
        ),
        col=str,
        fk=(StatuteRow.__tablename__, "id"),
    )

    @classmethod
    def list_affected_statutes(cls, c: Connection, pk: str) -> dict:
        sql_file = "statutes/list_affected_statutes.sql"
        results = c.db.execute_returning_dicts(
            corpus_sqlenv.get_template(sql_file).render(
                ref_tbl=cls.__tablename__,
                statute_tbl=StatuteRow.__tablename__,
                affecting_statute_id=pk,
            )
        )
        if results:
            return results[0]
        return {}

    @classmethod
    def list_affector_statutes(cls, c: Connection, pk: str) -> dict:
        sql_file = "statutes/list_affector_statutes.sql"
        results = c.db.execute_returning_dicts(
            corpus_sqlenv.get_template(sql_file).render(
                ref_tbl=cls.__tablename__,
                statute_tbl=StatuteRow.__tablename__,
                affected_statute_id=pk,
            )
        )
        if results:
            return results[0]
        return {}

    @classmethod
    def find_statute_in_unit(
        cls,
        text: str,
        mp: str,
        statute_id: str,
    ) -> Iterator["StatuteFoundInUnit"]:
        """Given text of a particular `material_path`, determine if there are
        statutes found by `get_statute_labels`; if they're found, determine
        the proper `StatuteFoundInUnit` to yield.
        """
        for rule in extract_rules(text):
            yield cls(
                material_path=mp,
                statute_id=statute_id,
                statute_category=rule.cat,
                statute_serial_id=rule.id,
                matching_statute_id=None,
            )

    @classmethod
    def extract_units(
        cls,
        pk: str,
        units: list["StatuteUnit"],
    ) -> Iterator["StatuteFoundInUnit"]:
        """Traverse the tree and search the caption and content of each unit
        for possible Statutes.
        """
        for u in units:
            if u.caption and u.content:
                text = f"{u.caption}. {u.content}"
                yield from cls.find_statute_in_unit(text, u.id, pk)
            elif u.content:
                yield from cls.find_statute_in_unit(u.content, u.id, pk)
            if u.units:
                yield from cls.extract_units(pk, u.units)

    @classmethod
    def get_statutes_from_references(cls, c: Connection) -> Iterator[dict]:
        """Extract relevant statute category and identifier pairs
        from the cls.__tablename__."""
        for row in c.db.execute_returning_dicts(
            corpus_sqlenv.get_template(
                "statutes/references/unique_statutes_list.sql"
            ).render(statute_references_tbl=cls.__tablename__)
        ):
            yield StatuteBase(**row).dict()

    @classmethod
    def update_statute_ids(cls, c: Connection) -> sqlite3.Cursor:
        """Since all statutes present in `db`, supply `matching_statute_id` in
        the references table."""
        with c.session as cur:
            return cur.execute(
                corpus_sqlenv.get_template("statutes/update_id.sql").render(
                    statute_tbl=StatuteRow.__tablename__,
                    target_tbl=cls.__tablename__,
                    target_col=cls.__fields__["matching_statute_id"].name,
                )
            )


class Statute(Integrator):
    id: str
    emails: list[EmailStr]
    meta: StatuteRow
    titles: list[StatuteTitleRow]
    tree: list[StatuteUnit]
    unit_fts: list[StatuteUnitSearch]
    material_paths: list[StatuteMaterialPath]
    statutes_found: list[StatuteFoundInUnit]

    @property
    def relations(self):
        return [
            (StatuteMaterialPath, self.material_paths),
            (StatuteUnitSearch, self.unit_fts),
            (StatuteTitleRow, self.titles),
            (StatuteFoundInUnit, self.statutes_found),
        ]

    @classmethod
    def from_page(cls, details_path: Path):
        # build and validate metadata from the path
        page = StatutePage.build(details_path)

        # assign row for creation
        meta = StatuteRow(**page.dict(exclude={"emails", "tree", "titles"}))

        # setup associated titles
        titles = [
            StatuteTitleRow(**statute_title.dict())
            for statute_title in page.titles
        ]

        # enable full text searches of contents of the tree; starts with `1.1.`
        fts = [
            StatuteUnitSearch(**unit)
            for unit in StatuteUnit.searchables(page.id, page.tree)
        ]

        # full text searches should includes a title row, i.e. node `1.``
        root_fts = [
            StatuteUnitSearch(
                statute_id=page.id,
                material_path="1.",
                unit_text=", ".join(
                    [f"{meta.statute_category} {meta.statute_serial_id}"]
                    + [t.text for t in titles]
                ),
            )
        ]

        return Statute(
            id=page.id,
            emails=page.emails,
            meta=meta,
            tree=page.tree,
            titles=titles,
            unit_fts=root_fts + fts,
            material_paths=[
                StatuteMaterialPath(**unit)
                for unit in StatuteUnit.granularize(
                    pk=page.id, nodes=page.tree
                )
            ],
            statutes_found=list(
                StatuteFoundInUnit.extract_units(pk=page.id, units=page.tree)
            ),
        )

    @classmethod
    def create_via_catid(cls, c: Connection, cat: str, id: str):
        """Create statute/s if the `cat` and `id` passed does not yet
        exist in the `statutes` table of the database."""
        if StatuteRow.get_id_via_catid(c, cat, id):
            return
        rule = Rule(cat=StatuteSerialCategory(cat), id=id)
        for folder in rule.extract_folders(STATUTE_PATH):
            obj = cls.from_page(folder / "details.yaml")
            idx = obj.insert_objects(c, StatuteRow, obj.relations)
            if idx:
                logger.debug(f"Added statute (cat/id): {idx}")

    @classmethod
    def make_tables(cls, c: Connection):
        """The bulk of the fields declared within the Statute
        container are table structures."""
        c.create_table(StatuteRow)  # corresponds to StatutePage
        c.create_table(StatuteTitleRow)  # corresponds to StatuteTitle
        c.create_table(StatuteUnitSearch)
        c.create_table(StatuteMaterialPath)
        c.create_table(StatuteFoundInUnit)
        c.db.index_foreign_keys()

    @classmethod
    def add_rows(cls, c: Connection):
        for detail in STATUTE_FILES:
            cat = detail.parent.parent.stem
            serial = detail.parent.stem
            try:
                obj = cls.from_page(detail)
                obj.insert_objects(c, StatuteRow, obj.relations)
                logger.debug(f"Added statute {cat=} {serial=} from {detail=}")
            except IntegrityError:
                logger.error(f"Already existing; skipped {cat=} {serial=}")
            except ValidationError as e:
                logger.error(f"Validation needed {detail=}; {e=}")
            except Exception as e:
                logger.error(f"Generic exception on insertion: {e=}")
        StatuteFoundInUnit.update_statute_ids(c)



from sqlite_utils.db import Table
from sqlpyd import Connection


def detect_statute_mp(code_pk: str, stat_id: str, tbl: Table, h: dict):
    mps, loc, cap, cont = (
        None,
        h.get("locator"),
        h.get("caption"),
        h.get("content"),
    )
    if all([loc, cap, cont, stat_id]):
        mps = list(
            tbl.rows_where(
                """codification_id = ? and locator = ?
                and caption = ? and content = ? and affector_statute_id = ?""",
                (code_pk, loc, cap, cont, stat_id),
                select="affector_material_path",
            )
        )
    elif all([loc, cap, stat_id]):
        mps = list(
            tbl.rows_where(
                """codification_id = ? and locator = ? and caption = ?
                and affector_statute_id = ?""",
                (code_pk, loc, cap, stat_id),
                select="affector_material_path",
            )
        )
    elif all([loc, cont, stat_id]):
        mps = list(
            tbl.rows_where(
                """codification_id = ? and locator = ? and content = ?
                and affector_statute_id = ?""",
                (code_pk, loc, cont, stat_id),
                select="affector_material_path",
            )
        )
    elif all([loc, stat_id]):
        mps = list(
            tbl.rows_where(
                """codification_id = ? and locator = ?
                and affector_statute_id = ?""",
                (code_pk, loc, stat_id),
                select="affector_material_path",
            )
        )
    if mps:
        if stat_mp := mps[0]["affector_material_path"]:
            return stat_mp
    return None


def detect_statute_id(code_pk: str, tbl: Table, h: dict):
    rows, stat, dt, vt = (
        None,
        h.get("statute"),
        h.get("date"),
        h.get("variant"),
    )
    if all([stat, dt, vt]):
        q = "codification_id = ? and statute = ? and date = ? and variant = ?"
        rows = list(
            tbl.rows_where(
                q,
                (code_pk, stat, dt, vt),
                select="affector_statute_id",
            )
        )
    elif all([stat, dt]):
        q = "codification_id = ? and statute = ? and date = ?"
        rows = list(
            tbl.rows_where(
                q,
                (code_pk, stat, dt),
                select="affector_statute_id",
            )
        )
    elif all([stat, vt]):
        q = "codification_id = ? and statute = ? and variant = ?"
        rows = list(
            tbl.rows_where(
                q,
                (code_pk, stat, vt),
                select="affector_statute_id",
            )
        )
    else:
        q = "codification_id = ? and statute = ?"
        rows = list(
            tbl.rows_where(
                q,
                (code_pk, stat),
                select="affector_statute_id",
            )
        )
    if rows:
        if stat_id := rows[0]["affector_statute_id"]:
            return stat_id
    return None


def detect_case_id(code_pk: str, tbl: Table, h: dict):
    if cite := h.get("citation"):
        if rows := list(
            tbl.rows_where(
                "codification_id = ? and citation = ?",
                (code_pk, cite),
                select="affector_decision_id",
            )
        ):
            return rows[0]["affector_decision_id"]


def set_histories(code_pk: str, nodes: list[dict], c: Connection):
    from corpus_base.decision import DecisionRow

    from corpus_x.codifications import CodeCitationEvent, CodeStatuteEvent
    from corpus_x.statutes import StatuteRow

    ct_evt = c.db[CodeCitationEvent.__tablename__]
    st_evt = c.db[CodeStatuteEvent.__tablename__]
    st = c.db[StatuteRow.__tablename__]
    dec = c.db[DecisionRow.__tablename__]
    for node in nodes:
        if h_list := node.get("history", None):
            for h in h_list:
                case_id = detect_case_id(code_pk, ct_evt, h)  # type: ignore
                if case_id:
                    case_date = dec.get(case_id)["date"]  # type: ignore
                    h["decision_id"] = case_id
                    h["decision_date"] = case_date

                stat_id = detect_statute_id(code_pk, st_evt, h)  # type: ignore
                if stat_id:
                    stat_date = st.get(stat_id)["date"]  # type: ignore
                    h["statute_id"] = stat_id
                    h["statute_date"] = stat_date
                    elements = (code_pk, stat_id, st_evt, h)
                    mp = detect_statute_mp(*elements)  # type: ignore
                    if mp:
                        h["statute_mp"] = mp
        if subunits := node.get("units"):
            set_histories(code_pk, subunits, c)



from pathlib import Path

from loguru import logger
from pydantic import Field
from sqlpyd import Connection, TableConfig
from statute_trees import DocPage, DocUnit, Page, generic_content, generic_mp

from .codifications import CodeRow
from .resources import DOCUMENT_PATH, Integrator


class DocRow(Page, TableConfig):
    __prefix__ = "lex"
    __tablename__ = "documents"


class DocUnitSearch(TableConfig):
    __prefix__ = "lex"
    __tablename__ = "document_fts_units"
    __indexes__ = [["document_id", "material_path"]]
    document_id: str = Field(..., col=str, fk=(DocRow.__tablename__, "id"))
    material_path: str = generic_mp
    unit_text: str = generic_content


class Document(Integrator):
    id: str
    meta: DocRow
    tree: list[DocUnit]
    unit_fts: list[DocUnitSearch]

    @property
    def relations(self):
        return [(DocUnitSearch, self.unit_fts)]

    @classmethod
    def make_tables(cls, c: Connection):
        if c.table(CodeRow):
            c.create_table(DocRow)
            c.create_table(DocUnitSearch)

    @classmethod
    def add_rows(cls, c: Connection):
        for f in DOCUMENT_PATH.glob("**/*.yaml"):
            obj = cls.from_page(f)
            idx = obj.insert_objects(c, DocRow, obj.relations)
            logger.debug(f"Added document {idx=} from {f.stem=}")

    @classmethod
    def from_page(cls, file_path: Path):
        page = DocPage.build(file_path)
        if not page:
            raise Exception(f"No page from {file_path=}")
        searachables = DocUnit.searchables(page.id, page.tree)
        return cls(
            id=page.id,
            emails=page.emails,
            meta=DocRow(**page.dict(exclude={"tree", "emails"})),
            tree=page.tree,
            unit_fts=[DocUnitSearch(**unit) for unit in searachables],
        )



import json
import sqlite3
from collections.abc import Iterator
from pathlib import Path

from corpus_base.decision import CitationRow, DecisionRow
from pydantic import EmailStr, Field
from sqlpyd import Connection, TableConfig
from statute_patterns import Rule, StatuteSerialCategory
from statute_patterns.components.utils import DETAILS_FILE
from statute_trees import (
    CitationAffector,
    CodePage,
    CodeUnit,
    Page,
    StatuteAffector,
    StatuteBase,
    generic_content,
    generic_mp,
)

from .resources import CODIFICATION_FILES, Integrator, corpus_sqlenv
from .statutes import Statute, StatuteMaterialPath, StatuteRow, StatuteTitleRow
from .utils import logger, set_histories


class CodeRow(Page, StatuteBase, TableConfig):
    __prefix__ = "lex"
    __tablename__ = "codifications"
    __indexes__ = [["statute_category", "statute_serial_id"]]

    @classmethod
    def get_id(cls, c: Connection, pk: str) -> str | None:
        tbl = c.table(cls)
        q = "id = ?"
        rows = list(tbl.rows_where(where=q, where_args=(pk,), select="id"))
        idx = rows[0]["id"] if rows else None
        return idx

    @classmethod
    def get_base_data(cls, c: Connection, pk: str) -> dict:
        sql_file = "codes/get_base.sql"
        template = corpus_sqlenv.get_template(sql_file)
        results = c.db.execute_returning_dicts(
            template.render(
                target_code_id=pk,
                code_tbl=cls.__tablename__,
                statute_tbl=StatuteRow.__tablename__,
                statute_title_tbl=StatuteTitleRow.__tablename__,
            )
        )
        if results:
            return results[0]
        return {}

    @classmethod
    def set_update_units(cls, c: Connection, pk: str) -> str:
        """Using data from codification events, add the `material_path`
        and `statute_id` of the affecting statute to each history node
        of the original `units` field."""
        tbl = c.db[cls.__tablename__]
        nodes = json.loads(tbl.get(pk)["units"])  # type: ignore
        set_histories(pk, nodes, c)  # recursive; updates nodes _in place_
        tbl.update(pk, {"units": json.dumps(nodes)})  # type: ignore
        return pk


class CodeUnitSearch(TableConfig):
    __prefix__ = "lex"
    __tablename__ = "codification_fts_units"
    __indexes__ = [["codification_id", "material_path"]]
    codification_id: str = Field(
        ..., col=str, fk=(CodeRow.__tablename__, "id")
    )
    material_path: str = generic_mp
    unit_text: str = generic_content


class CodeCitationEvent(CitationAffector, TableConfig):
    __prefix__ = "lex"
    __tablename__ = "codification_events_citation"
    __indexes__ = [["codification_id", "material_path"]]
    codification_id: str = Field(
        ..., col=str, fk=(CodeRow.__tablename__, "id")
    )
    material_path: str = generic_mp
    affector_decision_id: str = Field(
        None,
        description=(
            "The historical event is"
            " affected by a decision found in the decisions table."
        ),
        col=str,
        fk=(DecisionRow.__tablename__, "id"),
    )

    @classmethod
    def extract_units(cls, pk: str, units: list["CodeUnit"]):
        """Given a list of code units, extract affected units
        with their associated historical events."""
        for u in units:
            if u.history:
                for evt in u.history:
                    if isinstance(evt, CitationAffector):
                        yield CodeCitationEvent(
                            **evt.dict(),
                            material_path=u.id,
                            codification_id=pk,
                        )
            if u.units:
                yield from cls.extract_units(pk, u.units)

    @classmethod
    def update_decision_ids(cls, c: Connection):
        sql_file = "codes/events/set_decision_id.sql"
        template = corpus_sqlenv.get_template(sql_file)
        with c.session as cur:
            return cur.execute(
                template.render(
                    citation_tbl=CitationRow.__tablename__,
                    event_tbl=cls.__tablename__,
                )
            )


class CodeStatuteEvent(StatuteAffector, TableConfig):
    """Each `StatuteAffector` is a history node which takes into account
    a possible `date` and `variant` field to deal with duplicates.
    If they are present, these are used to detect the `affector_statute_id`.
    """

    __prefix__ = "lex"
    __tablename__ = "codification_events_statute"
    __indexes__ = [
        [
            "locator",
            "caption",
            "content",
            "statute_category",
            "statute_serial_id",
            "variant",
            "date",
        ],
        [
            "locator",
            "caption",
            "content",
            "statute_category",
            "statute_serial_id",
            "date",
        ],
        [
            "locator",
            "caption",
            "content",
            "statute_category",
            "statute_serial_id",
            "variant",
        ],
        [
            "locator",
            "caption",
            "content",
            "statute_category",
            "statute_serial_id",
        ],
        ["locator", "caption", "statute"],
        ["locator", "content", "statute"],
        ["locator", "statute"],
        ["codification_id", "material_path"],
        ["affector_statute_id", "affector_material_path"],
    ]
    codification_id: str = Field(
        ..., col=str, fk=(CodeRow.__tablename__, "id")
    )
    material_path: str = generic_mp
    affector_material_path: str | None = Field(None, col=str, index=True)
    affector_statute_unit_id: int = Field(
        None, col=str, fk=(StatuteMaterialPath.__tablename__, "id")
    )
    affector_statute_id: str | None = Field(
        None, col=str, fk=(StatuteRow.__tablename__, "id")
    )

    @classmethod
    def list_event_statutes(cls, c: Connection, pk: str) -> dict:
        sql_file = "codes/list_event_statutes.sql"
        template = corpus_sqlenv.get_template(sql_file)
        results = c.db.execute_returning_dicts(
            template.render(
                event_tbl=cls.__tablename__,
                statute_tbl=StatuteRow.__tablename__,
                target_id=pk,
            )
        )
        if results:
            return results[0]
        return {}

    @classmethod
    def extract_units(cls, pk: str, units: list["CodeUnit"]):
        """Given a list of code units, extract affected units
        with their associated historical events.
        """
        for u in units:
            if u.history:
                for evt in u.history:
                    if isinstance(evt, StatuteAffector):
                        yield cls(
                            **evt.dict(exclude_none=True),
                            material_path=u.id,
                            codification_id=pk,
                        )
            if u.units:
                yield from cls.extract_units(pk, u.units)

    @classmethod
    def get_statutes_from_events(cls, c: Connection) -> Iterator[dict]:
        """Extract relevant statute category and identifier pairs
        from the Codification Events table.
        """
        sql_file = "codes/list_unique_statute_category_idxes_from_events.sql"
        template = corpus_sqlenv.get_template(sql_file)
        q = template.render(statute_events_tbl=cls.__tablename__)
        for row in c.db.execute_returning_dicts(q):
            cat = row["statute_category"]
            idx = row["statute_serial_id"]
            if cat and idx:
                yield {"cat": cat, "id": idx}
            else:
                logger.error(f"Missing cat and idx in {row=}")
                continue

    @classmethod
    def add_statutes_from_events(cls, c: Connection):
        """Get `StatuteBases`; For each `StatuteBase`, extract possible
        folders. For each specific statute folder, setup a `Statute` object
        by extracting data from the folder and inserting the same into
        the database `db`."""
        for x in cls.get_statutes_from_events(c):
            rule = Rule(cat=StatuteSerialCategory(x["cat"]), id=x["id"])
            folders = rule.extract_folders()
            for folder in folders:
                content = folder / DETAILS_FILE
                detail = Rule.get_details(content)
                if not detail:
                    logger.error(f"Could not extract detail; {folder=}")
                    continue

                if is_existing_idx := StatuteRow.get_id(c, detail.id):
                    logger.debug(f"Existing statute: {is_existing_idx}")
                    continue

                else:
                    logger.debug(f"Attempt {rule=}; pk {folder.stem=}")
                    obj = Statute.from_page(content)
                    obj.insert_objects(c, StatuteRow, obj.relations)
                    logger.debug(f"Created statute: {obj}")

    @classmethod
    def update_statute_ids(cls, c: Connection) -> sqlite3.Cursor:
        """After running `cls.add_statutes_from_events()`, all Statutes
        contained in Codification statutory events will be present in
        the `db`. Supply the `affector_statute_id` of the
        CodeStatuteEvent table."""
        sql_file = "codes/events/set_statute_id.sql"
        template = corpus_sqlenv.get_template(sql_file)
        with c.session as cur:
            return cur.execute(
                template.render(
                    event_tbl=cls.__tablename__,
                    statute_tbl=StatuteRow.__tablename__,
                )
            )

    @classmethod
    def update_unit_ids(cls, c: Connection) -> sqlite3.Cursor:
        """After running `cls.update_affector_statute_ids()`, all statute
        affector ids have been updated. Supply the `affector_statute_unit_id`
        and  `affector_statute_unit_material_path` of the CodeStatuteEvent
        table."""
        sql_file = "codes/events/set_statute_unit_mp.sql"
        template = corpus_sqlenv.get_template(sql_file)
        with c.session as cur:
            return cur.execute(
                template.render(
                    event_tbl=cls.__tablename__,
                    mp_tbl=StatuteMaterialPath.__tablename__,
                )
            )

    @classmethod
    def fetch_unmaterialized(cls, c: Connection) -> list[dict] | None:
        """Search for unit events which were unable to get an
        affector material path."""
        sql_file = "codes/events/search_unmaterialized.sql"
        if rows := c.db.execute_returning_dicts(
            corpus_sqlenv.get_template(sql_file).render(
                event_tbl=cls.__tablename__,
            )
        ):
            return rows
        return None


class Codification(Integrator):
    id: str
    meta: CodeRow
    emails: list[EmailStr]
    tree: list[CodeUnit]
    unit_fts: list[CodeUnitSearch]
    stat_events: list[CodeStatuteEvent]
    cite_events: list[CodeCitationEvent]

    @property
    def relations(self):
        return [
            (CodeUnitSearch, self.unit_fts),
            (CodeStatuteEvent, self.stat_events),
            (CodeCitationEvent, self.cite_events),
        ]

    @classmethod
    def make_tables(cls, c: Connection):
        c.create_table(CodeRow)
        c.create_table(CodeUnitSearch)
        if c.table(StatuteRow):
            c.create_table(CodeStatuteEvent)
        if c.table(DecisionRow):
            c.create_table(CodeCitationEvent)

    @classmethod
    def add_rows(cls, c: Connection):
        # setup each codification row
        for raw_code in CODIFICATION_FILES:
            obj = cls.from_page(raw_code)
            idx = obj.insert_objects(c, CodeRow, obj.relations)
            logger.debug(f"Added codification {idx=} from {raw_code.stem=}")

        # add the base statute of codification by first transforming the table
        c.db[CodeRow.__tablename__].add_column(  # type: ignore
            col_name="statute_id",
            col_type=str,
            fk=StatuteRow.__tablename__,
            fk_col="id",
        )
        c.db.execute(
            corpus_sqlenv.get_template("codes/update_statute_id.sql").render(
                code_tbl=CodeRow.__tablename__,
                statutes_tbl=StatuteRow.__tablename__,
            )
        )

        # set events then update original trees
        CodeStatuteEvent.add_statutes_from_events(c)  # populate statutes table
        CodeStatuteEvent.update_statute_ids(c)  # needed to update tree
        CodeStatuteEvent.update_unit_ids(c)  # needed to update tree
        CodeCitationEvent.update_decision_ids(c)
        # update original trees
        [
            CodeRow.set_update_units(c, row["id"])
            for row in c.db[CodeRow.__tablename__].rows
        ]

    @classmethod
    def from_page(cls, file_path: Path):
        # build and validate metadata from the path
        page = CodePage.build(file_path)
        if not page:
            raise Exception(f"No page from {file_path=}")

        # assign row for creation
        meta = page.dict(exclude={"tree", "emails"})

        # enable full text searches of contents of the tree; starts with `1.1.`
        fts = [
            CodeUnitSearch(**unit)
            for unit in CodeUnit.searchables(pk=page.id, units=page.tree)
        ]

        # full text searches should includes a title row, i.e. node `1.``
        root_fts = [
            CodeUnitSearch(
                codification_id=page.id,
                material_path="1.",
                unit_text=", ".join(
                    [
                        f"{page.statute_category} {page.statute_serial_id}",
                        f"{page.title}",
                        f"{page.description}",
                    ]
                ),
            )
        ]

        return cls(
            id=page.id,
            emails=page.emails,
            meta=CodeRow(**meta),
            tree=page.tree,
            unit_fts=root_fts + fts,
            stat_events=list(
                CodeStatuteEvent.extract_units(pk=page.id, units=page.tree)
            ),
            cite_events=list(
                CodeCitationEvent.extract_units(pk=page.id, units=page.tree)
            ),
        )



import sqlite3
from collections.abc import Iterator
from typing import NamedTuple

import yaml
from citation_utils import Citation
from corpus_base import DECISION_PATH, CitationRow, DecisionRow, OpinionRow
from loguru import logger
from pydantic import Field
from rich.progress import track
from sqlpyd import Connection, TableConfig
from statute_patterns import Rule, StatuteSerialCategory, count_rules
from statute_patterns.components.utils import DETAILS_FILE
from statute_trees.resources import StatuteBase

from .resources import INCLUSION_FILE, STATUTE_PATH, corpus_sqlenv
from .statutes import StatuteRow


class StatuteInOpinion(StatuteBase, TableConfig):
    __prefix__ = "lex"
    __tablename__ = "opinion_statutes"
    __indexes__ = [["statute_category", "statute_serial_id"]]
    opinion_id: str = Field(..., col=str, fk=(OpinionRow.__tablename__, "id"))
    included_statute_id: str | None = Field(
        None,
        description=(
            "This will be initially absent but"
            "will be updateable through a later update process."
        ),
        col=str,
        fk=(StatuteRow.__tablename__, "id"),
    )  # note difference of statute_id contained here vs statutes.StatuteFK
    mentions: int = Field(
        description=(
            "Each opinion can contain a list of statutes"
            " and their corresponding number."
        ),
        col=int,
    )

    @classmethod
    def extracted(cls, op_id: str, text: str) -> Iterator["StatuteInOpinion"]:
        try:
            for counted in count_rules(text):
                yield cls(
                    opinion_id=op_id,
                    statute_category=StatuteSerialCategory(counted["cat"]),
                    statute_serial_id=counted["id"],
                    mentions=counted["mentions"],
                    included_statute_id=None,
                )
        except Exception as e:
            logger.error(f"Bad statute detection; {op_id=}; {e=}")

    @classmethod
    def most_popular(cls, c: Connection) -> list[dict]:
        """Get a list of unique statutes included during
        `Inclusion.insert_objs_to_db()` and order them according to
        their popularity."""
        template_name = "decisions/inclusions/popular_statutes.sql"
        template = corpus_sqlenv.get_template(template_name)
        return c.db.execute_returning_dicts(
            template.render(op_stat_tbl=cls.__tablename__)
        )

    @classmethod
    def update_statute_ids(cls, c: Connection) -> list[dict]:
        """Assuming proper inserts of missing statutes from
        cls.add_statutes(), update the rows with their proper foreign keys."""
        return c.db.execute(
            corpus_sqlenv.get_template("statutes/update_id.sql").render(
                statute_tbl=StatuteRow.__tablename__,
                target_tbl=cls.__tablename__,
                target_col=cls.__fields__["included_statute_id"].name,
            )
        )

    @classmethod
    def add_statutes(cls, c: Connection):
        """When `Inclusion.from_files_to_db()` first creates StatuteInOpinion
        rows, these rows do not include a `statute_id` (see `opinion_statutes`
        table).

        This function adds statutes to the database from the local repository
        based on the "most popular" StatuteInOpinion rows; since the
        statute_ids now exist, they can be referenced.
        """
        from corpus_x import Statute

        for i in StatuteInOpinion.most_popular(c):
            rule = Rule(cat=StatuteSerialCategory(i["cat"]), id=i["idx"])
            for folder in rule.extract_folders(STATUTE_PATH):
                content_file = folder / DETAILS_FILE
                detail = Rule.get_details(content_file)
                if not detail:
                    logger.error(f"Could not extract detail; {folder=}")
                    continue
                try:
                    obj = Statute.from_page(content_file)
                    idx = obj.insert_objects(c, StatuteRow, obj.relations)
                    if idx:
                        logger.debug(f"Created statute: {idx}")
                except Exception as e:
                    logger.error(f"Did not make statute {content_file=}; {e=}")
                    continue


class CitationInOpinion(Citation, TableConfig):
    __prefix__ = "lex"
    __tablename__ = "opinion_citations"
    __indexes__ = [
        ["included_decision_id", "scra"],
        ["included_decision_id", "phil"],
        ["included_decision_id", "docket"],
        ["included_decision_id", "offg"],
    ]
    opinion_id: str = Field(..., col=str, fk=(OpinionRow.__tablename__, "id"))
    included_decision_id: str | None = Field(
        None,
        description=(
            "This will be initially absent but will be"
            " updateable through a later update process."
        ),
        col=str,
        fk=(DecisionRow.__tablename__, "id"),
    )

    @classmethod
    def extracted(cls, op_id: str, text: str) -> Iterator["CitationInOpinion"]:
        try:
            base = dict(opinion_id=op_id, included_decision_id=None)
            for cite in Citation.extract_citations(text):
                data: dict = cite.dict()
                res = data | base
                yield cls(**res)  # type: ignore
        except Exception as e:
            logger.error(f"Bad citations; {op_id=}; {e=}")

    @classmethod
    def most_popular(cls, c: Connection) -> list[dict]:
        """Get a list of unique citations included during
        `Inclusion.insert_objs_to_db()` and order them
        according to their popularity.
        """
        template_name = "decisions/inclusions/popular_citations.sql"
        template = corpus_sqlenv.get_template(template_name)
        return c.db.execute_returning_dicts(
            template.render(
                cite_tbl=CitationRow.__tablename__,
                op_cite_tbl=cls.__tablename__,
            )
        )

    @classmethod
    def update_decision_ids(cls, c: Connection) -> sqlite3.Cursor:
        """When CitationInOpinions rows are first added by
        `Inclusion.insert_objs_to_db()`, they lack a decision.
        Since the Decision IDs already exist, can update the
        CitationInOpinion rows.
        """
        template_name = "decisions/inclusions/update_decision_id.sql"
        template = corpus_sqlenv.get_template(template_name)
        return c.db.execute(
            template.render(
                cite_tbl=CitationRow.__tablename__,
                target_tbl=cls.__tablename__,
                target_col=cls.__fields__["included_decision_id"].name,
            )
        )


class Inclusion(NamedTuple):
    """Is not necessary as a table since this is used as a namespace to
    collect related entries and consolidating them into separate files.
    """

    source: str  # whether sc / legacy
    origin: str  # identifying folder
    decision_id: str  # source of the opinion
    opinion_id: str  # source of the text
    text: str  # text to examine
    statutes: list[StatuteInOpinion]
    citations: list[CitationInOpinion]

    @classmethod
    def get_base_data(cls, c: Connection, pk: str) -> dict:
        sql_file = "decisions/get_base.sql"
        template = corpus_sqlenv.get_template(sql_file)
        results = c.db.execute_returning_dicts(
            template.render(
                target_decision_id=pk,
                decision_tbl=DecisionRow.__tablename__,
            )
        )
        if results:
            return results[0]
        return {}

    @classmethod
    def list_opinions_of_decision(cls, c: Connection, pk: str) -> dict:
        sql_file = "decisions/list_opinions_of_decision.sql"
        template = corpus_sqlenv.get_template(sql_file)
        results = c.db.execute_returning_dicts(
            template.render(
                target_decision_id=pk,
                opinion_tbl=OpinionRow.__tablename__,
                op_cite_tbl=CitationInOpinion.__tablename__,
                decision_tbl=DecisionRow.__tablename__,
                op_stat_tbl=StatuteInOpinion.__tablename__,
                statute_tbl=StatuteRow.__tablename__,
            )
        )
        if results:
            return results[0]
        return {}

    @classmethod
    def make_tables(cls, c: Connection):
        """These tables reference foreign keys.

        Presumed to already exist:

        1. Statute table
        2. Decision table

        Args:
            c (Connection): sqlpyd connection to the database
        """
        if c.table(StatuteRow):
            c.create_table(StatuteInOpinion)
        if c.table(DecisionRow):
            c.create_table(CitationInOpinion)

    @property
    def content_for_file(self):
        msg = f"Inclusions detected in {self.path_to_folder=}"
        statutes = [i.dict(exclude_none=True) for i in self.statutes]
        citations = [i.dict(exclude_none=True) for i in self.citations]
        if not statutes and not citations:
            logger.debug(f"No {msg.lower()}")
            return
        else:
            logger.debug(msg)
            return {"statutes": statutes, "citations": citations}

    @property
    def path_to_folder(self):
        folder = DECISION_PATH / self.source / self.origin
        if not folder.exists():
            logger.error(f"Bad {folder=} stored in the database.")
        return folder


def populate_db_with_inclusions(c: Connection):
    """Assuming that `create_inclusion_files_from_db_opinions()` has
    previously run, we can extract the contents of each file and insert
    them into the database.
    """
    paths = DECISION_PATH.glob(f"**/{INCLUSION_FILE}")
    for path in track(paths, description="Pull inclusions..."):
        obj = yaml.safe_load(path.read_bytes())
        if obj.get("statutes"):
            c.add_records(StatuteInOpinion, obj["statutes"])
        if obj.get("citations"):
            c.add_records(CitationInOpinion, obj["citations"])


def create_inclusion_files_from_db_opinions(c: Connection):
    """Need a connection to the database to retrieve Opinion objects.

    From the text found in each opinion, extract statutes and citations
    and save to an inclusion file.

    The inclusion file shall be stored in in the same source
    repository (see source / origin fields)."""

    def read_opinions(c: Connection) -> list[dict[str, str]]:
        """Join each opinion of each decision together to collect
        all opinions in the database.

        Each collected entry will consist of:

        1. `source` and `origin` to help with the getting the local path.
        2. `decision_id` and `opinion_id` to create the resulting record.
        3. `text` of the opinion which will be used to determine inclusions.
        """
        sql_file = "decisions/inclusions/read_opinions.sql"
        return c.db.execute_returning_dicts(
            corpus_sqlenv.get_template(sql_file).render(
                opinion_tbl=OpinionRow.__tablename__,
                decision_tbl=DecisionRow.__tablename__,
            )
        )

    def set_inclusion_objects(rows: list[dict]):
        """An `Inclusion` instance is a NamedTuple which consists of
        included statutes and decisions from an opinion's text.
        This function helps extract such "inclusions"."""
        for o in track(rows, description="Set inclusions..."):
            obj = Inclusion(
                **o,
                statutes=list(
                    StatuteInOpinion.extracted(
                        op_id=o["opinion_id"],
                        text=o["text"],
                    )
                ),
                citations=list(
                    CitationInOpinion.extracted(
                        op_id=o["opinion_id"],
                        text=o["text"],
                    )
                ),
            )
            if obj.content_for_file:
                yield obj

    opinions: list[dict] = read_opinions(c)
    inclusions: Iterator[Inclusion] = set_inclusion_objects(opinions)
    for obj in inclusions:
        logger.debug(f"Creating {obj.source=} / {obj.origin=}")
        f = DECISION_PATH / obj.source / obj.origin / INCLUSION_FILE
        f.unlink(missing_ok=True)  # replace
        with open(f, "w") as writefile:
            yaml.safe_dump(obj.content_for_file, writefile)


def set_inclusions(c: Connection):
    """Collect the pre-processed data and insert the same
    into the created database tables.

    Estimate at the end of 2022 (factors to consider):

    1. the last time data was scraped as raw files,
    2. the time separate opinions were manually included

    Note: the statute and inclusion tables need to be created
    before the pre-processed data can be inserted.

    Result:

    As of end of 2022:

    table | row count | purpose
    --:|:-- |:--
    `CitationsInOpinions` | ~484k | itemize citations found per opinion
    `StatutesInOpinions` | ~99k  | itemize statutes found per opinion

    After `populate_db_with_inclusions` is run, what will
    exists in the database are records of statutes / citations but not
    the statute_ids / decision_ids themselves. See `purpose` noted above
    for the citations and statutes found per opinion.

    Recall that the inclusion files are created because of `corpus-base`
    tables. At the time the `corpus-base` tables were created, there were
    were no StatuteInOpinions and CitationInOpinions tables yet since
    these only exist in `corpus-x`

    An update will need to be made so that these tables get a foreign key
    to their proper tables.

    Args:
        c (Connection): sqlpyd connection to the database
    """
    populate_db_with_inclusions(c)
    StatuteInOpinion.update_statute_ids(c)
    CitationInOpinion.update_decision_ids(c)
    c.db.index_foreign_keys()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['corpus_x', 'corpus_x.utils']

package_data = \
{'': ['*'],
 'corpus_x': ['sql/analysis/*',
              'sql/base/*',
              'sql/codes/*',
              'sql/codes/events/*',
              'sql/decisions/*',
              'sql/decisions/inclusions/*',
              'sql/statutes/*',
              'sql/statutes/references/*']}

install_requires = \
['corpus-base>=0.1.10,<0.2.0', 'statute-trees>=0.1.0,<0.2.0']

setup_kwargs = {
    'name': 'corpus-x',
    'version': '0.0.2',
    'description': 'Add codification and statute tables to pre-existing corpus-base database.',
    'long_description': '# corpus-x\n\n![Github CI](https://github.com/justmars/corpus-x/actions/workflows/main.yml/badge.svg)\n\nCreate the `x.db` sqlite database for lawdata; utilized in the [LawSQL dataset](https://lawsql.com).\n\n## Documentation\n\nSee [documentation](https://justmars.github.io/corpus-x), building on top of [corpus-base](https://justmars.github.io/corpus-base)\n\n## Development\n\nCheckout code, create a new virtual environment:\n\n```sh\npoetry add corpus-x # python -m pip install corpus-x\npoetry update # install dependencies\npoetry shell\n```\n\nRun tests:\n\n```sh\npytest\n```\n',
    'author': 'Marcelino G. Veloso III',
    'author_email': 'mars@veloso.one',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://lawsql.com',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.11,<4.0',
}


setup(**setup_kwargs)



__version__ = "0.0.2"

from .__main__ import (
    build_x_tables,
    get_codification,
    get_decision,
    get_document,
    get_statute,
    reset_x,
    setup_x,
)
from .codifications import Codification
from .documents import Document
from .inclusions import (
    CitationInOpinion,
    Inclusion,
    StatuteInOpinion,
    set_inclusions,
)
from .statutes import Statute
